/*                         
 *                         DIRECT CONDENSE 2.0
 *                         
 * dcgeneric/dccheckvec.c   Tool to cross check orbit databases.
 * 
 * Copyright (C) 2000 Frank Luebeck and Max Neunhoeffer, 
 *                    Lehrstuhl D fuer Mathematik, RWTH Aachen
 *                    
 * This file is protected under the GNU General Public License, 
 * (see the file "GPL.txt" in the main distribution directory for details). 
 * 
 * This program is for looking up points from one orbit database in the 
 * database of another hash server already running.
 * 
 *  $Id: dccheckvec.c,v 1.3 2000/04/28 09:26:20 neunhoef Exp $
 */


#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>

#include "clntserv.h"
#include "dcutil.h"
#include "dcservices.h"

int main(int argc, char *argv[])
{
  FILE *f;
  long hashserv;
  long veclength;
  long first,last;
  CS_BUFTYPE question;
  CS_BUFTYPE answer;
  long i,j;
  long todo;
  long *p;

  if (argc != 7) {
    printf("Usage:  dccheckvec vec_file veclength hashservhost hashservport "
           "first last\n");
    exit(1);
  }
  if ((f = fopen(argv[1], "r")) == NULL) {
    printf("Can't open input file: %s.\n", argv[1]);
    exit(1);
  }
  hashserv = cs_connect(argv[3],atoi(argv[4]));
  if (hashserv < 0) {
    printf("Can't connect to hash server: %s:%d\n",argv[3],atoi(argv[4]));
    exit(2);
  }
  veclength = atol(argv[2]);
  first = atol(argv[5]);
  last = atol(argv[6]);
  fseek(f,veclength * first,SEEK_SET);
  question = cs_new_buffer(100 * veclength);
  answer = cs_new_buffer(100 * sizeof(long));

  for (i = first;i <= last;) {
    if (last - i + 1 > 100)
      todo = 100;
    else
      todo = last - i + 1;

    if ((ssize_t) fread(question.buf,veclength,todo,f) < todo) {
      printf("Can't read vectors!\n");
      exit(3);
    }

    question.len = todo*veclength;

    if (cs_request(hashserv,HASHQUERY,question,&answer) < 0) {
      printf("Error in request to hash server!\n");
      exit(4);
    }

    for (j = 0,p = answer.buf;j < todo;j++,p++) {
      if ((i+j - first) % 10 == 0)
        printf("\n%8ld: ",i+j);

      if (*p == HTFAIL)
        printf("???? ");
      else
        printf("%4ld ",*p);
    }
    i += todo;
  }

  printf("\n");
  fclose(f);
  return 0;
}

