/*                         
 *                         DIRECT CONDENSE 2.0
 *                         
 * dcgeneric/dcclient.c   Generic part of the calculation client(s).
 * 
 * Copyright (C) 2000 Frank Luebeck and Max Neunhoeffer, 
 *                    Lehrstuhl D fuer Mathematik, RWTH Aachen
 *                    
 * This file is protected under the GNU General Public License, 
 * (see the file "GPL.txt" in the main distribution directory for details). 
 * 
 * This program does the main work for the calculation. It connects to the
 * job and hash servers. The actual implementation of the operation is in
 * other subdirectories.
 * 
 * $Id: dcclient.c,v 1.11 2000/04/28 09:26:46 neunhoef Exp $
 */

/* The following changes the behaviour of custom.h : */
#define DC_ISCLIENT

#include "clntserv.h"
#include "dcutil.h"
#include "dcservices.h"
#include "custom.h"

#ifdef SHOWTIMINGS
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#endif

/* global variables which must be set in the initialiation */
long dc_length_pt;   /* length of point in orbit in bytes */
long dc_ng;          /* number of elements to condense (in startup number
                        of given generators of G */
void **ggen;         /* list of dc_ng void pointers to generators of G */
long dc_nk;          /* number of given generators of K */
void **kgen;         /* list of dc_nk void pointers to generators of K */
void *start;         /* pointer to start point */
long htklen;         /* length of hash table for K-orbits */
long maxminpart;     /* maximal number of minimal elements in a part the 
			partitions of a K-orbit */
long maxpart;        /* maximal number of elements in a part */
long maxnrpartskorb; /* maximal number of parts in a K-orbit */
long maxextra;       /* maximal extra number of minimal representatives per
			K-orbit */
long dc_max_fact_req; /* this times nr of hash servers requests to hash 
                         servers are produced at most in `CondMatRows' */
 
/* more global variables */
long nrhashserv;     /* number of hash servers */
long *hashservers;   /* pointer to list of hash servers */
HashType *htk;       /* hash table for a K-orbit (no values, see `main' */
CS_BUFTYPE *quest;   /* list of question buffers, one for each hash server */
/* //CS_BUFTYPE *answ;   list of answer buffers, one for each hash server  */
CS_BUFTYPE kanswer;  /* for sending K orbits to hash servers */
void *minbuf;        /* pointer to big enough space to hold set of minimal 
                        elements in a part of the partition of the orbits */
void *partbuf1;  /* pointer to big enough spaces to hold a part of 
                        the partition of the orbits */
void *repsbuf;        /* pointer to big enough space to hold representatives
                         of K-orbit */
void *extrabuf;       /* pointer to big enough space to hold extra minimals
                         of K-orbit */
void *pt1, *pt2;      /* pointers to big enough space to hold a point of 
                         the orbit */


/* This orbit computes representatives and all minimal elements for 
   the partition of a K-orbit. It also sends the information to the hash
   servers. */
void KOrbitByPartition(void *pt)
{
  long i, j, k, l, h, nr, len, nreps, nextra, korblen;
  void *ptr, *pr, *preps, *pextra;
#ifdef SHOWTIMINGS
  struct rusage usebefore,useafter;

  getrusage(RUSAGE_SELF, &usebefore);
#endif

  ClearHT(htk);
  pextra = extrabuf;

  /* Add minimals of part of `pt' */
  DC_FINDALLMINIMAL(pt, minbuf, &nr);
  /* we collect representatives for parts in `repsbuf' and further
     minimal points in `extrabuf' */
  nreps = 1;
  memcpy(repsbuf, minbuf, dc_length_pt);
  nextra = nr - 1;
  if (nr>1) {
    memcpy(pextra,(char *)minbuf+dc_length_pt, (nr-1)*dc_length_pt);
    pextra = (char *)pextra + (nr-1)*dc_length_pt;
  }
  AddHT(htk, repsbuf, dc_length_pt, 0);
  for (i=1, ptr=extrabuf; i<nr; ptr = (char *) ptr + dc_length_pt, i++) 
    AddHT(htk, ptr, dc_length_pt, 0);
  
  /* loop over parts */
  korblen = 0;
  for (i=0, preps=repsbuf; i<nreps; preps = (char *)preps+dc_length_pt, i++) {
    /* evaluate part */
    DC_FINDPART(preps, partbuf1, &len);
    korblen += len;
    /*   // printf("U%ld:%ld ", i, len);fflush(stdout); */
    /* loop over points in part */
    for (j=0, pr=partbuf1; j<len; pr = (char *) pr + dc_length_pt, j++) {
      /* loop over generators of K */
      for (k=0; k<dc_nk; k++) {
        DC_OPERATION(pr, kgen[k], pt1);
        DC_FINDMINIMAL(pt1, pt2);
        h = ValueHT(htk, pt2, dc_length_pt);
        if (h==HTFAIL) {   /* new part found, compute info */
          /* similar code as for part of starting point */
          DC_FINDALLMINIMAL(pt2, minbuf, &nr);
          memcpy((char *)repsbuf+dc_length_pt*nreps, minbuf, dc_length_pt);
          AddHT(htk, (char *)repsbuf+dc_length_pt*nreps, dc_length_pt, 0);
          nreps++;
          nextra += (nr - 1);
          if (nr>1) {
            memcpy(pextra, (char *)minbuf+dc_length_pt, (nr-1)*dc_length_pt);
          }
          for (l=1, ptr=pextra; l<nr; ptr = (char *) ptr + dc_length_pt, l++) 
            AddHT(htk, ptr, dc_length_pt, 0);
          pextra = (char *) pextra + (nr-1)*dc_length_pt;
        } 
      }
    }
  }

  /******   now send result to hash servers     *******/
  /* first split into questions for the different servers */
  len = (nreps + nextra)*dc_length_pt + 4*sizeof(long);
  for (i=0; i<nrhashserv; i++) {
    quest[i] = cs_new_buffer(len);
    /*  //    cs_ensure_bufferlength(&(quest[i]), len); */
    quest[i].len = 3*sizeof(long); /* header: orblen, nrreps, nrextra */
    *(((long*)quest[i].buf)) = korblen;
    *(((long*)quest[i].buf)+1) = 0;
    *(((long*)quest[i].buf)+2) = 0;
  }
  /* split list of representatives */
  for (j=0, preps=repsbuf; j<nreps; j++, preps = (char *) preps+dc_length_pt) {
    i = DC_HASHSERVFUN(preps);
    (*(((long*)quest[i].buf)+1))++;
    memcpy((char *) (quest[i].buf) + quest[i].len, preps, dc_length_pt);
    quest[i].len += dc_length_pt;
  }
  /* split list of extra minimals */
  for (j=0, pextra=extrabuf; j<nextra;
       j++, pextra = (char *) pextra + dc_length_pt) {
    i = DC_HASHSERVFUN(pextra);
    (*(((long*)quest[i].buf)+2))++;
    memcpy((char *) (quest[i].buf)+quest[i].len, pextra, dc_length_pt);
    quest[i].len += dc_length_pt;
  }  
  /* find first non empty list */
  for (j=0; quest[j].len == 3*sizeof(long); j++);
  cs_request(hashservers[j], HASHNEW, quest[j], &kanswer);
  /* send the others if orbit is new, it has  number `nr' */
  if ((nr = *((long *)kanswer.buf)) > 0) { 
    for (i=0; i<nrhashserv; i++) {
      if (i != j) {
	      printf("send result #%ld to server %ld (%ld)\n",nr,i,(long) quest[i].len);fflush(stdout);
        memcpy((char *)(quest[i].buf) + quest[i].len, 
	       (void*)(&nr), sizeof(long));
        quest[i].len += sizeof(long);
        cs_request(hashservers[i], HASHNEWNR, quest[i], &kanswer);
      }
    }
  }
  /* free buffers allocated in `cs_new_buffer' */
  for (i=0; i<nrhashserv; free(quest[i].buf), i++); 
  printf("Computed and sent K-orbit %ld.\n", nr); fflush(stdout);
#ifdef SHOWTIMINGS
  getrusage(RUSAGE_SELF, &useafter);
  printf("Time used for this K-orbit: %ld (user) %ld (system) [ms]\n",
          (useafter.ru_utime.tv_sec-usebefore.ru_utime.tv_sec) * 1000 
         + (useafter.ru_utime.tv_usec - useafter.ru_utime.tv_usec),
          (useafter.ru_stime.tv_sec-usebefore.ru_stime.tv_sec) * 1000 
         + (useafter.ru_stime.tv_usec - useafter.ru_stime.tv_usec));
  fflush(stdout);
#endif
}


/* The central function for computing the result of a job. */

void CondMatRows(CS_BUFTYPE reps, LongList *crows)
{
  void *ptr;        /* pointer to representatives of parts */
  long i, j, k;
  long nrreqs=0;    /* nr of current requests to hashservers */     
  long *lp;         /* for answer numbers from hashservers */
  long len, a;
  long *count;
  long rand;/* a random number */
  CS_REQTYPE *r;    /* pointer to request type, to handle answers
                       from hash servers */

  CS_BUFTYPE *quest; /* pointer to list of question buffers */
  long **priv;      /* data for requests: how many points are mapped under
                       given element in G */

  /* allocate space for request buffers */
  quest = (CS_BUFTYPE*)malloc(nrhashserv*sizeof(CS_BUFTYPE));
  /*  // answ = (CS_BUFTYPE*)malloc(nrhashserv*sizeof(CS_BUFTYPE)); */
  priv = (long**)malloc(nrhashserv*sizeof(long *));

  /* loop over representatives of parts */
  for (ptr=reps.buf; (char *) ptr < (char *) reps.buf+reps.len || nrreqs>0; ) {
    /* first we handle finished requests to the hash servers */
    while ((r=cs_whichready(-1))) {
      /*  //  lp = (long *)(r->answer.buf);   */
      len = (r->answer.len)/sizeof(long);
      /* the random number just in case the data are sorted somehow */
      rand=(long)((unsigned long)random() % len);
      /*   // printf("rand: %ld/%ld\n", rand,len);fflush(stdout); */
      for (i=rand, lp = ((long *)(r->answer.buf))+rand; 
	   i<len && *lp != HTFAIL; i++, lp++);
      if (i==len) for (i=0, lp = (long *)(r->answer.buf); 
		       i<rand && *lp != HTFAIL; i++, lp++);
      if (*lp != HTFAIL) { /* all K-orbits of vectors in request were known */
        /* adjust `crows' according to results */
        lp = (long *)r->answer.buf;
        for (j=0, count = (long*)r->private; j<dc_ng; j++, count++) {
          for (k=0; k<(*count); lp++, k++) {
            a = *lp;
            AppendZerosLongList(crows+j, a);
            (crows[j].list[a-1])++;
          }
        }
        /* free memory and throw out request */
        free(r->private);
	/*  //        free(r->answer->buf);   */
        /*  //        free(r->question.buf);  */
        cs_forget(r, 1); nrreqs--;
      }
      else { /* new K-orbit found, compute its minimal elements and report
                to hash servers */
        /* stop waiting requests, since their answers will quite 
           probably also contain HTFAIL's */
        cs_stopall_nb(-1); 
        /* We will have requests to hash servers, sending new orbit.
           So we try now to get current requests finished. */ 
        cs_do_io();
        /* compute K-orbit and send data to hash servers */
        KOrbitByPartition((char *) (r->question.buf) + i*dc_length_pt);
        /* Now all requests in queue are finished or not yet handled.
           We reset all of them for (re)sending. */
        cs_repeatall_nb(-1);
        cs_continueall_nb(-1);
      }
    }

    /* Now compute new requests, if there are not too many of them already 
       in the queue. */
    if ((char *) ptr < (char *) (reps.buf) + reps.len && 
	nrreqs < dc_max_fact_req * nrhashserv) {
      /* evaluate part */
      DC_FINDPART(ptr, partbuf1, &len);
      /* allocate buffers */
      for (i=0; i<nrhashserv; i++) {
        quest[i] = cs_new_buffer(len*dc_length_pt*dc_ng);
	/*  //        answ[i] = cs_new_buffer(len*dc_ng*sizeof(long));  */
        priv[i] = (long *)calloc(dc_ng,sizeof(long));
      }
      /* compute minimals and sort into right requests */
      for (j=0; j<dc_ng; j++) {  /* loop over elements to condense */
	for (k=0; k<len; k++) {   /* loop over points in part */
	  /*  //pt1 = partbuf1 + k*dc_length_pt; */
	  DC_OPERATION((char *) partbuf1 + k*dc_length_pt, ggen[j], pt1);
	  DC_FINDMINIMAL(pt1, pt2);
	  /* find corresponding hash server and copy  pt2 into the request */
	  i = DC_HASHSERVFUN(pt2);
	  memcpy((char *) (quest[i].buf) + quest[i].len, pt2, dc_length_pt);
	  quest[i].len += dc_length_pt;
	  priv[i][j]++;
	}
      }
      for (i=0; i<nrhashserv; i++) {
	if (quest[i].len > 0) {
	  cs_request_nb_malloc(hashservers[i], HASHQUERY, quest[i], 
			       cs_new_buffer(0L), (void*)(priv[i]));
	  nrreqs++;
	}
	else {
	  free(priv[i]);
	  free(quest[i].buf);
	}
      }
      /* goto next point */
      ptr = (char *) ptr + dc_length_pt;
    }
  }
  free(quest);
  free(priv);
}

/* print usage info */
void usage()
{
  printf("\nUsage:\n  dcclient input jobserv port nrhashsvrs hashserv1 port ");
  printf("hashserv2 port ...\n"); fflush(stdout);
  exit(1);
}

int main(int argc, char *argv[])
{
  FILE *input;               /* file of input data for initialization */
  long jobserv, jobnr;              /* job server */
  long i, sum, *ptr;
  CS_BUFTYPE question, answer, reps; /* buffers for requests */
  LongList *crows;           /* pointer to list of result rows */

  if (argc == 1) usage();

  /* file with input data for the application, nothing about its
     format is defined here, the file is an argument to the `dcclient_init'
     function */
  if (( input = fopen(argv[1], "r")) == NULL) {
    usage();
    perror( "Cannot open input file:");
    exit(1);
  } 

  /* initialization must be explained in custom.h/.c 
     it must assign certain global variables, as stated in the 
     beginning of this file */
  DC_CLIENT_INIT(input);

  if (argc < 6) exit(0);  /* to test initialization */
 
  /* job server */
  jobserv = cs_connect(argv[2], atoi(argv[3]));
  
  /* all hash servers */
  nrhashserv = atol(argv[4]);
  hashservers = (long *)malloc(nrhashserv * sizeof(long));
  for (i=0; i<nrhashserv; i++)
    hashservers[i] = cs_connect(argv[2*i+5], atoi(argv[2*i+6]));

  /* global hash table for K-orbits */
  htk = NewHTNoVals(htklen, DC_HASHFUN, DC_CMP_PTS);
  
  /* question buffers for requests to hash servers */
  quest = (CS_BUFTYPE *)malloc(nrhashserv*sizeof(CS_BUFTYPE));
  /*  //  answ = (CS_BUFTYPE *)malloc(nrhashserv*sizeof(CS_BUFTYPE)); */

  /* answer buffer for sending new K-orbits */
  kanswer = cs_new_buffer(sizeof(long));
  
  /* allocating some more global buffers, see their declaration */
  minbuf = (void*)calloc(maxminpart*dc_length_pt, sizeof(char));
  partbuf1 = (void*)calloc(maxpart*dc_length_pt, sizeof(char));
  /*  //partbuf2 = (void*)calloc(maxpart*dc_length_pt, sizeof(char)); */
  repsbuf = (void*)calloc(maxnrpartskorb*dc_length_pt, sizeof(char));
  extrabuf = (void*)calloc(maxextra*dc_length_pt, sizeof(char));
  pt1 = (void*)calloc(dc_length_pt, sizeof(char));
  pt2 = (void*)calloc(dc_length_pt, sizeof(char));

  /* initialize some more buffers */
  question = cs_new_buffer(sizeof(long));
  answer = cs_new_buffer(sizeof(long));
  reps = cs_new_buffer(sizeof(long));
  crows = (LongList *)malloc(dc_ng * sizeof(LongList));
  for (i=0; i<dc_ng; i++) crows[i] = *(NewLongList(1000));

  /* main loop: get new jobs and work */
  for (;;) {
    question.len = 0;
    cs_request(jobserv, NEWJOB, question, &answer);
    jobnr = *((long *)answer.buf);
    if (jobnr == 0) { 
      /* orbit of start point, this function also distribute information
         to hash servers */
      KOrbitByPartition(start);
    }
    else if (jobnr == -2)
      sleep(5); /* we ask again in 5 seconds */
    else if (jobnr == -1) {
      printf("No more jobs - bye.\n"); fflush(stdout);
      exit(1);
    }
    else { /* the standard case */
      /* collect representatives of K-orbit from hash servers */

      srandom(jobnr);
      *((long *)question.buf) = jobnr;
      question.len = sizeof(long);
      for (i=0, reps.len=0; i<nrhashserv; i++) {
        /* get representatives of K-orbit from hash server i and store
           them in `reps' */
        while (1) {
          cs_request_malloc(hashservers[i], HASHORB, question, &answer);
          if (*((long *)answer.buf)<0) { /* orbit not yet known */
            printf("Waiting for representatives from hash server.\n"); 
	    fflush(stdout); 
            sleep(5);
          }
          else
            break;
        } 
        cs_ensure_bufferlength(&reps, reps.len + answer.len - sizeof(long));
          memcpy((char *) reps.buf + reps.len, 
		 (char *) answer.buf + sizeof(long), 
                 answer.len-sizeof(long));
          reps.len += (answer.len - sizeof(long));
      }
      /* the main routine computes the rows of the condensation matrices
         and stores them in `crows' */
      for (i=0; i<dc_ng; crows[i].len = 0, i++); /* reset lengths of crow's */
      printf("Condensing now orbit #%ld.\n",jobnr); fflush(stdout);
      CondMatRows(reps, crows);
      /* send the result to the job server */
      for (sum=2+dc_ng, i=0; i<dc_ng; sum+=crows[i].len, i++);
      cs_ensure_bufferlength(&question, sum * sizeof(long));
      question.len = sum * sizeof(long);
      ptr = (long *)question.buf;
      *(ptr++) = jobnr;
      *(ptr++) = dc_ng;
      for (i=0; i<dc_ng; i++) {
        *(ptr++) = crows[i].len;
        memcpy((void *)ptr, (void *)crows[i].list, 
                 crows[i].len * sizeof(long));
        ptr += crows[i].len;
      }
      printf("Sending result rows for job number %ld.\n", jobnr); 
      fflush(stdout);
      cs_request(jobserv, GETRESULT, question, &answer);
    }
  }
}
