/*                         
 *                         DIRECT CONDENSE 2.0
 *                         
 * mtx_vec_U_quo/custom.h   Headers for operation matrices on vectors
 * 
 * Copyright (C) 2000 Frank Luebeck and Max Neunhoeffer, 
 *                    Lehrstuhl D fuer Mathematik, RWTH Aachen
 *                    
 * This file is protected under the GNU General Public License, 
 * (see the file "GPL.txt" in the main distribution directory for details). 
 * 
 * In this file the spezialization to matrices acting on vectors
 * by right multiplication using the C-MeatAxe is implemented. The
 * "U-quotient"-trick is used to handle vector recognition by
 * enumerating orbits "by U-orbits". 
 * 
 *  $Id: custom.h,v 1.3 2000/04/28 12:27:14 neunhoef Exp $
 */

/* include files for the application */
#include <string.h>
#include <stdio.h>

#include "dcutil.h"

#include "meataxe.h"

/* define's here */


/* Some of the declarations of functions below may be  substituted by 
   macros for some applications */


/* void DC_CLIENT_INIT(FILE *input) 
   this function must set at least the following global variables for the
   client, see the beginning of `dcclientgeneric.c':
   
long dc_length_pt;      length of point in orbit in bytes   
long dc_ng;             number of elements to condense (in startup number
                        of given generators of G   
void **ggen;            list of dc_ng void pointers to generators of G   
long dc_nk;             number of given generators of K   
void **kgen;            list of dc_nk void pointers to generators of K   
void *start;            pointer to start point
long htklen;            length of hash table for K-orbits   
long maxminpart;        maximal number of minimal elements in a part the 
			partitions of a K-orbit
long maxpart;           maximal number of elements in a part
long maxnrpartskorb;    maximal number of parts in a K-orbit
long maxextra;          maximal extra number of minimal representatives per
			K-orbit 
long dc_max_fact_req;   this times nr of hash servers requests to hash 
                        servers are produced at most in `CondMatRows' 

Furthermore it should set up all global data needed by the other functions
defined below.

*/
extern long dc_length_pt;
#ifdef DC_ISCLIENT
extern long dc_ng;
extern void **ggen;
extern long dc_nk;
extern void **kgen;
extern void *start;
extern long htklen;
extern long maxminpart;
extern long maxpart;
extern long maxnrpartskorb;
extern long maxextra;
extern long dc_max_fact_req;

extern long nrhashserv;     /* number of hash servers */
#endif
/* ifdef DC_ISCLIENT */

void DC_CLIENT_INIT(FILE *input);


/* void DC_OPERATION(void* pt, void *g, void *img)
   This should compute the image of point pt under group element g and 
   store the resulting point in img. Points and group elements are referred 
   to by void pointers, img must point to a large enough part of the memory 
   to hold the result.   
*/

void DC_OPERATION(void* pt, void *g, void *img);


/* int DC_CMP_PTS(void* pt1, void* pt2, size_t len) 
   This is for comparing two points (of data length `len'), returns 
   0 in case of equality. 
*/
int DC_CMP_PTS(void *pt1, void *pt2, size_t len);

/* long DC_HASHFUN(void *pt, size_t len);
   The hash function, it gets a point as one argument and the length of 
   its data, it has to return a long number.
*/
long DC_HASHFUN(void *pt, size_t len);


/* long DC_HASHSERVFUN(void *);
   The hash server function, it gets a point as one argument and has
   to return the number of a hash server in the range from
   0..#hashservers-1: */
long DC_HASHSERVFUN(void *pt);

/* void DC_FINDMINIMAL(void *pt1, void *pt2);
   This function is called with a point pt1 and has to store one 
   minimal point in the same part of the K-orbit into pt2. 
   pt2 must point to a large enough part of the memory.
*/
void DC_FINDMINIMAL(void *pt1, void *pt2);

/* void DC_FINDALLMINIMAL(void *pt1, void *ptr, long *nr);
   This function is called with a point and has to store all 
   minimal points in the same part of the K-orbit into ptr. 
   ptr must point to a large enough part of the memory. *nr will
   be assigned to the number of found minimal points.
*/
void DC_FINDALLMINIMAL(void *pt1, void *ptr, long *nr);

/* void DC_FINDPART(void *pt1, void *ptr, long *nr);
   This function is called with a point and has to store all 
   points in the same part of the K-orbit into ptr. 
   ptr must point to a large enough part of the memory. *nr will
   be assigned to the number of found  points.
*/
void DC_FINDPART(void *pt1, void *ptr, long *nr);

/* void DC_HASHSERVINIT(LongList *l);
   This function  is called with some data in a long list to configure 
   some global variables in a hash server. */
void DC_HASHSERVINIT(LongList *l);

