/*                         
 *                         DIRECT CONDENSE 2.0
 *                         
 * perms_S21_8841/customhash.c   Implementation of operation of S21 for hashs.
 * 
 * Copyright (C) 2000 Frank Luebeck and Max Neunhoeffer, 
 *                    Lehrstuhl D fuer Mathematik, RWTH Aachen
 *                    
 * This file is protected under the GNU General Public License, 
 * (see the file "GPL.txt" in the main distribution directory for details). 
 * 
 * This file contains the spezialization of the operation to the case 
 * of S21 operating in its natural permutation representation. The 
 * condensation subgroup is a 8841 Young subgroup. Here are the routines
 * for the hash server.
 * 
 *  $Id: customhash.c,v 1.4 2000/04/28 13:36:18 neunhoef Exp $
 *
 */


/* include files for the application */
#include "custom.h"

/* define's here */

/* extra variables */
long permdegree;        /* degree of permutation group G */
void *minptpart;        /* pointer to a point used in DC_FINDPART */
 
/* Some of the declarations of functions below may be  substituted by 
   macros for some applications */

/* void DC_OPERATION(void* pt, void *g, void *img)
   This should compute the image of point pt under group element g and 
   store the resulting point in img. Points and group elements are referred 
   to by void pointers, img must point to a large enough part of the memory 
   to hold the result.   
*/
void DC_OPERATION(void* pt, void *g, void *img)
{
  long i;
  int *gg;
  char *p, *pi;
  
  gg = (int*)g;
  p = (char*) pt + sizeof(long);
  pi = (char*) img + sizeof(long);
  for (i=0; i<permdegree; pi[gg[i]] = p[i], i++);
}

/* long DC_CMP_PTS(void* pt1, void* pt2, size_t len) 
   This is for comparing two points (of data length `len'), returns 
   0 in case of equality. 

   done by  ???????????????????????
   #define DC_CMP_PTS memcmp
*/
int DC_CMP_PTS(void *pt1, void *pt2, size_t len)
{
  return memcmp(pt1, pt2, len);
}

/* long DC_HASHFUN(void *pt, size_t len);
   The hash function, it gets a point as one argument and the length of 
   its data, it has to return a long number.
*/
long DC_HASHFUN(void *pt, size_t len)
{
  len = len; /* unused */
  /* very easy here, since the hashvalue is the beginning of the data */
  return *((long*)pt);
}

/* void DC_HASHSERVINIT(LongList *l);
   This function  is called with some data in a long list to configure 
   some global variables in a hash server. */
void DC_HASHSERVINIT(LongList *l)
{
  dc_length_pt = l->list[1];
}


