/*                         
 *                         DIRECT CONDENSE 2.0
 *                         
 * test/testclntnb.c   Test program for the clntserv library, client part.
 * 
 * Copyright (C) 2000 Frank Luebeck and Max Neunhoeffer, 
 *                    Lehrstuhl D fuer Mathematik, RWTH Aachen
 *                    
 * This file is protected under the GNU General Public License, 
 * (see the file "GPL.txt" in the main distribution directory for details). 
 * 
 * Test program for the clntserv library, client part. Uses non-blocking
 * I/O-routines.
 * 
 * $Id: testclntnb.c,v 1.2 2000/04/28 13:36:18 neunhoef Exp $
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "clntserv.h"

int main(int argc, char *argv[])
{
  unsigned short int port;
  long server1,server2;
  int sleeptime;
  CS_BUFTYPE buf1,buf2;
  CS_REQTYPE *req1,*req2;
  CS_REQTYPE *req;
  int res;

  if (argc != 6) {
    puts("Usage: testclnt host1 port1 host2 port2 sleeptime");
    exit(1);
  }

  sleeptime = atoi(argv[5]);

  port = atoi(argv[2]);
  server1 = cs_connect(argv[1],port);

  port = atoi(argv[4]);
  server2 = cs_connect(argv[3],port);

  buf1 = cs_new_buffer(sizeof(long));
  buf2 = cs_new_buffer(sizeof(long));

  while (1) {
    /* we send away two requests: */
    buf1.len = 0;
    req1 = cs_request_nb(server1,123,buf1,buf1,NULL);
    buf2.len = 0;
    req2 = cs_request_nb(server2,123,buf2,buf2,NULL);

    /* Now we wait for both answers: */
    while (req1 || req2) {
      req = cs_whichready(-1);
      
      if (req) {
	if (req == req1)
	  printf("whichready said #1!\n");
	if (req == req2)
	  printf("whichready said #2!\n");
      }

      if (req1) {
	res = cs_isready(req1);
	if (res > 0) {
	  printf("Got number: %ld Server 1\n",*((long *)(req1->answer.buf)));
	  fflush(stdout);
	  cs_forget(req1,0);
	  req1 = NULL;
	} else if (res < 0) {
	  fprintf(stderr,"Error with request to Server 1.\n");
	  cs_forget(req1,0);
	  req1 = NULL;
	}
      }

      if (req2) {
	res = cs_isready(req2);
	if (res > 0) {
	  printf("Got number: %ld Server 2\n",*((long *)(req2->answer.buf)));
	  fflush(stdout);
	  cs_forget(req2,0);
	  req2 = NULL;
	} else if (res < 0) {
	  fprintf(stderr,"Error with request to Server 2.\n");
	  cs_forget(req2,0);
	  req2 = NULL;
	}
      }
    }

    sleep(sleeptime);
  }
    
  return 0;
}
